/*
 * Copyright (c) Atmosphère-NX
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stratosphere.hpp>
#include "impl/diag_print_debug_string.hpp"

namespace ams::diag {

    namespace {

        inline NORETURN void AbortWithValue(u64 debug) {
            #if defined(ATMOSPHERE_BOARD_NINTENDO_NX)
            /* Just perform a data abort. */
            register u64 addr __asm__("x27") = FatalErrorContext::StdAbortMagicAddress;
            register u64 val __asm__("x28")  = FatalErrorContext::StdAbortMagicValue;
            while (true) {
                __asm__ __volatile__ (
                    "mov x0, %[debug]\n"
                    "str %[val], [%[addr]]\n"
                    :
                    : [debug]"r"(debug), [val]"r"(val), [addr]"r"(addr)
                    : "x0"
                );
            }
            #else
            AMS_UNUSED(debug);
            std::abort();
            #endif
            __builtin_unreachable();
        }

        inline void DebugLog(const char *format, ...) __attribute__((format(printf, 1, 2)));

#ifdef AMS_ENABLE_DETAILED_ASSERTIONS
        constinit os::SdkRecursiveMutex g_debug_log_lock;
        constinit char g_debug_buffer[0x400];

        void DebugLogImpl(const char *format, ::std::va_list vl) {
            std::scoped_lock lk(g_debug_log_lock);

            util::VSNPrintf(g_debug_buffer, sizeof(g_debug_buffer), format, vl);

            diag::impl::PrintDebugString(g_debug_buffer, strlen(g_debug_buffer));
        }

        void DebugLog(const char *format, ...) {
            ::std::va_list vl;
            va_start(vl, format);
            DebugLogImpl(format, vl);
            va_end(vl);
        }

#else
        void DebugLog(const char *format, ...)  { AMS_UNUSED(format); }
#endif

    }

    NORETURN NOINLINE WEAK_SYMBOL void AssertionFailureImpl(const char *file, int line, const char *func, const char *expr, u64 value, const char *format, ...) {
        #if defined(ATMOSPHERE_OS_HORIZON)
        DebugLog("%016" PRIx64 ": Assertion Failure\n", os::GetCurrentProgramId().value);
        #else
        DebugLog("0100000000000000: Assertion Failure\n");
        #endif
        DebugLog("        Location:   %s:%d\n", file, line);
        DebugLog("        Function:   %s\n", func);
        DebugLog("        Expression: %s\n", expr);
        DebugLog("        Value:      %016" PRIx64 "\n", value);
        DebugLog("\n");
#ifdef AMS_ENABLE_DETAILED_ASSERTIONS
        {
            ::std::va_list vl;
            va_start(vl, format);
            DebugLogImpl(format, vl);
            va_end(vl);
        }
#else
        AMS_UNUSED(format);
#endif
        DebugLog("\n");

        AbortWithValue(value);
    }

    NORETURN NOINLINE WEAK_SYMBOL void AssertionFailureImpl(const char *file, int line, const char *func, const char *expr, u64 value) {
        #if defined(ATMOSPHERE_OS_HORIZON)
        DebugLog("%016" PRIx64 ": Assertion Failure\n", os::GetCurrentProgramId().value);
        #else
        DebugLog("0100000000000000: Assertion Failure\n");
        #endif
        DebugLog("        Location:   %s:%d\n", file, line);
        DebugLog("        Function:   %s\n", func);
        DebugLog("        Expression: %s\n", expr);
        DebugLog("        Value:      %016" PRIx64 "\n", value);
        DebugLog("\n");
        DebugLog("\n");

        AbortWithValue(value);
    }

    NORETURN NOINLINE WEAK_SYMBOL void AbortImpl(const char *file, int line, const char *func, const char *expr, u64 value, const char *format, ...) {
        #if defined(ATMOSPHERE_OS_HORIZON)
        DebugLog("%016" PRIx64 ": Abort Called\n", os::GetCurrentProgramId().value);
        #else
        DebugLog("0100000000000000: Abort Called\n");
        #endif
        DebugLog("        Location:   %s:%d\n", file, line);
        DebugLog("        Function:   %s\n", func);
        DebugLog("        Expression: %s\n", expr);
        DebugLog("        Value:      %016" PRIx64 "\n", value);
        DebugLog("\n");
#ifdef AMS_ENABLE_DETAILED_ASSERTIONS
        {
            ::std::va_list vl;
            va_start(vl, format);
            DebugLogImpl(format, vl);
            va_end(vl);
        }
#else
        AMS_UNUSED(format);
#endif
        DebugLog("\n");

        AbortWithValue(value);
    }

    NORETURN NOINLINE WEAK_SYMBOL void AbortImpl(const char *file, int line, const char *func, const char *expr, u64 value) {
        #if defined(ATMOSPHERE_OS_HORIZON)
        DebugLog("%016" PRIx64 ": Abort Called\n", os::GetCurrentProgramId().value);
        #else
        DebugLog("0100000000000000: Abort Called\n");
        #endif
        DebugLog("        Location:   %s:%d\n", file, line);
        DebugLog("        Function:   %s\n", func);
        DebugLog("        Expression: %s\n", expr);
        DebugLog("        Value:      %016" PRIx64 "\n", value);
        DebugLog("\n");
        DebugLog("\n");

        AbortWithValue(value);
    }

    NORETURN NOINLINE WEAK_SYMBOL void AbortImpl() {
        AbortWithValue(0);
    }

}
